/*
  GetColor.cpp

  Implementation of the GetColor function
*/

#include <stdio.h>
#include <math.h>

#include "TGAFile.h"

/*
void GetColor(const int x,
              const int y,
              RGBQUAD* pColor)

Retrieve the color of a pixel.
To retrieve the color of a pixel directly from the TGA file, see GetColorFromFile().

Parameters:
x                      const int       x coordinate of pixel requested
y                      const int       y coordinate of pixel requested
pColor                 RGBQUAD*        Address of target byte triple

Return value: none

Global variables affected: none
*/
void TGAFile::GetColor(const int x, const int y, RGBQUAD* pColor)
{
  if(!m_bInMemory)
  {
    GetColorFromFile(x, y, pColor);
  }
  else
  {
    if((x >= 0) && (x < m_pHeader->image_width) && (y >= 0) && (y < m_pHeader->image_height))
    {
      int i;
      unsigned char* pPos1;
      unsigned char* pPos2;

      pPos1 = m_pData + y * m_dwWidthDWORDAligned + x * m_nBytesPerPixel;
      pPos2 = (unsigned char *)pColor;
      /* Put into pColor. */
      for(i = 0; i < m_nBytesPerPixel; i++)
      {
        *pPos2++ = *pPos1++; /* BLUE, GREEN, RED [, ALPHA]  */
      }
    }
#ifdef _DEBUG
    else
    {
      fprintf(stderr,
              "Error: Overflow in TGAFile::GetColor(%d, %d) - file size is only (%d, %d).\n",
              x, y, m_pHeader->image_width, m_pHeader->image_height);
    }
#endif
  }
}

/*
void GetColor(const unsigned int x,
              const unsigned int y,
              RGBQUAD* pColor)

Retrieve the color of a pixel.
To retrieve the color of a pixel directly from the TGA file, see GetColorFromFile().

Parameters:
x                      const int       x coordinate of pixel requested
y                      const int       y coordinate of pixel requested
pColor                 RGBQUAD*        Address of target byte triple

Return value: none

Global variables affected: none
*/
void TGAFile::GetColor(const unsigned int x, const unsigned int y, RGBQUAD* pColor)
{
	GetColor((int)(x), (int)(y), pColor);
}

/*
void GetColorFromFile(const int x,
                      const int y,
                      RGBQUAD* pColor)

Retrieve the color of a pixel directly from the TGA file.

Parameters:
x                      const int       x coordinate of pixel requested
y                      const int       y coordinate of pixel requested
pColor                 RGBQUAD*        Address of target byte triple

Return value: none

Global variables affected: none
*/
void TGAFile::GetColorFromFile(const int x, const int y, RGBQUAD* pColor)
{
  if((x >= 0) && (x < m_pHeader->image_width) && (y >= 0) && (y < m_pHeader->image_height))
  {
    fseek(m_pInputFile,
          (long)sizeof(TGAFILEHEADER) +
            (long)m_pHeader->ident_len +
            ((long)y * (long)m_pHeader->image_width +
            (long)x) * 3,
          SEEK_SET);
    fread(pColor, 3, 1, m_pInputFile);
  }
#ifdef _DEBUG
  else
  {
    fprintf(stderr,
            "Error: Overflow in TGAFile::GetColorFromFile(%d, %d) - file size is only (%d, %d).\n",
            x, y, m_pHeader->image_width, m_pHeader->image_height);
  }
#endif
}

/*
void GetColor(double x,
              double y,
              RGBQUAD* pColor)

retrieve the color of a pixel

Parameters:
x                      double          x coordinate of pixel requested
y                      double          y coordinate of pixel requested
pColor                 RGBQUAD*        Address of target byte triple

Return value: none

Global variables affected: none
*/
void TGAFile::GetColor(double x, double y, RGBQUAD* pColor)
{
  if(x < 0.0)
  {
    x = 0.0;
  }
  if(x > m_pHeader->image_width - 2.0)
  {
    x = m_pHeader->image_width - 2.0;
  }
  if(y < 0.0)
  {
    y = 0.0;
  }
  if(y > m_pHeader->image_height - 2.0)
  {
    y = m_pHeader->image_height - 2.0;
  }

  int xint = (int)x;
  int yint = (int)y;
  double xrest   = x - floor(x);
  double yrest   = y - floor(y);
  double x1_rest = 1.0 - xrest;
  double y1_rest = 1.0 - yrest;

  RGBQUAD Color[4];

  GetColor(xint,     yint,     Color);
  GetColor(xint + 1, yint,     Color + 1);
  GetColor(xint,     yint + 1, Color + 2);
  GetColor(xint + 1, yint + 1, Color + 3);
  int nBlue   = (int)(yrest * (x1_rest * Color[2].rgbBlue + xrest * Color[3].rgbBlue) + y1_rest * (x1_rest * Color[0].rgbBlue + xrest * Color[1].rgbBlue));
  int nGreen  = (int)(yrest * (x1_rest * Color[2].rgbGreen + xrest * Color[3].rgbGreen) + y1_rest * (x1_rest * Color[0].rgbGreen + xrest * Color[1].rgbGreen));
  int nRed    = (int)(yrest * (x1_rest * Color[2].rgbRed + xrest * Color[3].rgbRed) + y1_rest * (x1_rest * Color[0].rgbRed + xrest * Color[1].rgbRed));
  pColor->rgbBlue  = nBlue;
  pColor->rgbGreen = nGreen;
  pColor->rgbRed   = nRed;
  pColor->rgbReserved = 0;
}
