/*
  Open.cpp

  Implementation of the Open function.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>

#include "TGAFile.h"

/*
int Open(const char *szFilename)

Open TGA file

Parameters:
szFilename           const char*     Name of input file

Return value:
0 if no error occurred
a non-zero value if an error occurred
*/

int TGAFile::Open(const char *szFilename)
{
  unsigned long lWidthInBytes;
  unsigned char *pPixels;
  size_t lResult = 0L;
  int y;

  if(*szFilename == 0)
  {
    fprintf(stderr, "TGAFile::Open(): Empty filename.\n");
    return 8;
  }
  if(!strcmp(szFilename, "-"))
  {
    // The filename is "-" ... this is stdin.
    // Set "stdin" to have binary mode:
    int nResult = _setmode(_fileno(stdin), _O_BINARY);
    if(nResult == -1)
    {
      fprintf(stderr, "TGAFile::Open(): Cannot set mode _O_BINARY for stdin.\n");
      return 1;
    }
    m_pInputFile = stdin;
  }
  else
  {
    m_pInputFile = fopen(szFilename, "rb");
    if(!m_pInputFile)
    {
      fprintf(stderr, "TGAFile::Open(): Cannot open file \"%s\".\n", szFilename);
      return 2;
    }
  }
  m_pHeader = new TGAFILEHEADER;
  if(m_pHeader == 0)
  {
    fprintf(stderr, "TGAFile::Open(): Cannot allocate memory for TGAFILEHEADER.\n");
    return 3;
  }

  if(fread(m_pHeader,
           1,
           sizeof(TGAFILEHEADER),
           m_pInputFile) != sizeof(TGAFILEHEADER))
  {
    fprintf(stderr, "TGAFile::Open(): Cannot read TGAFILEHEADER.\n");
    delete m_pHeader;
    m_pHeader = 0;
    return 4;
  }
  if(m_pHeader->cmap_type != nopalette ||
     m_pHeader->image_type != uncompr_rgb)
  {
    fprintf(stderr, "TGAFile::Open(): TGA file is not uncompressed RGB.\n");
    delete m_pHeader;
    m_pHeader = 0;
    return 5;
  }

  /*
    Allocate memory for DIB
    DIB lines are DWORD aligned in memory
  */

  m_nBytesPerPixel = m_pHeader->pixel_size / 8;
  lWidthInBytes = m_nBytesPerPixel * m_pHeader->image_width;
  m_dwWidthDWORDAligned = ((lWidthInBytes - 1) / 4 + 1) * 4;

  if(m_bInMemory || !strcmp(szFilename, "-"))
  {
    m_pData = new unsigned char[m_dwWidthDWORDAligned * m_pHeader->image_height];

    if(m_pData == 0)
    {
      fprintf(stderr, "TGAFile::Open(): Cannot allocate memory for TGA data.\n");
      delete m_pHeader;
      m_pHeader = 0;
      return 6;
    }

    /* Go read the bits. */

    lResult = 0L;
    if(m_pHeader->image_discr.OrgBit)
    {
      // The OrgBit is set, see TGAFILEHEADER definition.
      // Read the first line of data into the highest line of the image.
      pPixels = m_pData + (m_pHeader->image_height - 1) * m_dwWidthDWORDAligned;

      // Read top down.
      for(y = 0; y < m_pHeader->image_height; y++)
      {
        lResult = fread(pPixels,
                        1,
                        lWidthInBytes,
                        m_pInputFile);
        pPixels -= m_dwWidthDWORDAligned;
      }
    }
    else
    {
      // The OrgBit is cleared, see TGAFILEHEADER definition.
      // Read the first line of data into the lowest line of the image.
      pPixels = m_pData;

      // Read bottom up.
      for(y = 0; y < m_pHeader->image_height; y++)
      {
        lResult = fread(pPixels,
                        1,
                        lWidthInBytes,
                        m_pInputFile);
        pPixels += m_dwWidthDWORDAligned;
      }
    }
  }

  if(!strcmp(szFilename, "-"))
  {
    // The filename is "-" ... this is stdin.
    m_pInputFile = 0;
  }
  else
  {
    if(m_bInMemory)
    {
      fclose(m_pInputFile);
      m_pInputFile = 0;
      if(lResult != lWidthInBytes)
      {
        fprintf(stderr, "TGAFile::Open(): Error reading TGA file.\n");
        delete m_pData;
        m_pData = 0;
        delete m_pHeader;
        m_pHeader = 0;
        if(m_pInputFile)
        {
          fclose(m_pInputFile);
          m_pInputFile = 0;
        }
        return 7;
      }
    }
  }
  return 0;
}
