#include "StdAfx.h"

#include <afxtempl.h>
#include "HTTPRequest.h"
#include "geoDataClient.h"
#include "TimeoutDialog.h"

#include <list>
#include "Waypoint.h"
#include "geoDataClientDialog.h"

const int BLOCKSIZE = 2048;

CHTTPRequest::CHTTPRequest(CListBox &messages,
                           int &maxExtent,
                           bool &bC,
                           CProgressCtrl &pC,
                           CList<CString,CString&> &pIT,
                           int nHTTPTimeout) :
  messagesListBox(messages),
  maxListboxTextExtent(maxExtent),
  bCancel(bC),
  progressControl(pC),
  progressInformationText(pIT)
{
  data.Empty();
  outputFilename.Empty();
  this->nHTTPTimeout = nHTTPTimeout;
}

CHTTPRequest::~CHTTPRequest(void)
{
}

// returns 0 if OK.
int CHTTPRequest::execute(const CString &strOriginalUrl,
                          const CString &suggestedOutputFilename,
                          bool bMemoryOnly)
{
/*
Your goal               Actions you take                      Effects
-------------------------------------------------------------------------------------------
Begin an HTTP           Create a CInternetSession object.     Initializes WinInet and
session.                                                      connects to server.

Connect to an           Use                                   Returns a CHttpConnection
HTTP server.            CInternetSession::GetHttpConnection.  object.

Open an HTTP request.   Use CHttpConnection::OpenRequest.     Returns a CHttpFile object.

Send an HTTP request.   Use CHttpFile::AddRequestHeaders and  Finds the file.
                        CHttpFile::SendRequest.               Returns FALSE if the file
                                                              is not found.

Read from the file.     Use CHttpFile.                        Reads the specified number
                                                              of bytes using a buffer you
                                                              supply.

Handle exceptions.      Use the CInternetException class.     Handles all common Internet
                                                              exception types.

End the HTTP session.   Dispose of the CInternetSession       Automatically cleans up open
                        object.                               file handles and connections.
-------------------------------------------------------------------------------------------
*/

  CString strUrl = strOriginalUrl;
  DWORD dwRet = 0;

  while(true) // only exit via break/return.
  {
    CString strServerName;
    CString strObject;
    INTERNET_PORT nPort = 0;
    DWORD dwServiceType = 0;
    CHttpConnection *pHttpServer = NULL;
    CHttpFile *pHttpFile = NULL;
    if(!AfxParseURL(strUrl, dwServiceType, strServerName, strObject, nPort))
    {
      CString message;
      message.Format(_T("Error: can only use URLs beginning with http://"));
      addMessage(message);
      return 1;
    }
    if(dwServiceType != INTERNET_SERVICE_HTTP)
    {
      CString message;
      message.Format(_T("Error: can only use URLs beginning with http://"));
      addMessage(message);
      return 2;
    }
    addMessage(_T("Opening Internet..."));
    CInternetSession session(_T("TruFlite's geoDataClient"),   // pstrAgent
                             1);                               // dwContext

    bool retry = true;
    while(retry)
    {
      TRY
      {
        session.SetOption(INTERNET_OPTION_RECEIVE_TIMEOUT, nHTTPTimeout * 1000);
        pHttpServer = session.GetHttpConnection(strServerName, nPort);
        LPCTSTR ppstrAcceptTypes[] =
        {
          "text/*",
          "application/zip",
          "*/*",
          NULL
        };
        pHttpFile =
          pHttpServer->OpenRequest(CHttpConnection::HTTP_VERB_GET,  // nVerb
                                  strObject,                       // target object
                                  NULL,                            // referer
                                  1,                               // dwContext
                                  ppstrAcceptTypes,                // content types accepted
                                  NULL,                            // HTTP version: "HTTP/1.0"
                                  INTERNET_FLAG_NO_AUTO_REDIRECT); // redirections should not be automatically handled in CHttpFile::SendRequest
        pHttpFile->AddRequestHeaders(_T("Accept: text/*\r\n"
                                        "Accept: application/zip\r\n"
                                        "Accept: */*\r\n"
                                        "User-Agent: Mozilla\r\n"));
        pHttpFile->SendRequest();
        pHttpFile->QueryInfoStatusCode(dwRet);
      }
      CATCH(CInternetException, pEx)
      {
        // Simply show an error message to the user.
        CTimeoutDialog timeoutDialog(0, nHTTPTimeout);
        INT_PTR result = timeoutDialog.DoModal();
        nHTTPTimeout = timeoutDialog.getHTTPTimeout();
        if(result == IDYES)
        {
          retry = true;
        }
        else
        {
          retry = false;
        }
        dwRet = HTTP_STATUS_REQUEST_TIMEOUT;  // This is somewhat inconsistent, but very expedient.
      }
      END_CATCH
      if(dwRet != HTTP_STATUS_REQUEST_TIMEOUT)
      {
        retry = false;  // dwRet was set by anything but the CATCH block.
      }
    }
    // if access was denied, prompt the user for the password
    while(dwRet == HTTP_STATUS_DENIED)
    {
      DWORD dwPrompt;
      dwPrompt =
        pHttpFile->ErrorDlg(NULL,
                            ERROR_INTERNET_INCORRECT_PASSWORD,
                            FLAGS_ERROR_UI_FLAGS_GENERATE_DATA | FLAGS_ERROR_UI_FLAGS_CHANGE_OPTIONS,
                            NULL);

      // if the user cancelled the dialog, bail out

      if(dwPrompt != ERROR_INTERNET_FORCE_RETRY)
      {
        CString message;
        message.Format(_T("Access denied: Invalid password"));
        addMessage(message);
        pHttpFile->Close();
        delete pHttpFile;
        pHttpFile = 0;
        pHttpServer->Close();
        delete pHttpServer;
        pHttpServer = 0;
        return 1;
      }
      pHttpFile->SendRequest();
      pHttpFile->QueryInfoStatusCode(dwRet);
    }
    // were we redirected?
    // these response status codes come from WININET.H

    CString message;
    message.Format(_T("Status Code is %lu"), dwRet);
    addMessage(message);
    switch(dwRet)
    {
      case HTTP_STATUS_MOVED:
      case HTTP_STATUS_REDIRECT:
      case HTTP_STATUS_REDIRECT_METHOD:
      {
        CString strNewLocation;
        int nWaitTime = 0;
        pHttpFile->QueryInfo(HTTP_QUERY_RAW_HEADERS_CRLF, strNewLocation);
        int nPlace = strNewLocation.Find(_T("Refresh: "));
        if(nPlace != -1)
        {
          strNewLocation = strNewLocation.Mid(nPlace + 9);
          nPlace = strNewLocation.Find('\n');
          if(nPlace > 0)
          {
            strNewLocation = strNewLocation.Left(nPlace);
          }
          nPlace = strNewLocation.Find('\r');
          if(nPlace > 0)
          {
            strNewLocation = strNewLocation.Left(nPlace);
          }
        }
        else
        {
          strNewLocation.Empty();
        }
        nPlace = strNewLocation.Find(_T(";"));
        if(nPlace > 0)
        {
          CString strWaitTime = strNewLocation.Left(nPlace);
          nWaitTime = atoi(strWaitTime);
        }

        pHttpFile->QueryInfo(HTTP_QUERY_RAW_HEADERS_CRLF, strNewLocation);
        nPlace = strNewLocation.Find(_T("Location: "));
        if(nPlace == -1)
        {
          CString message;
          message.Format(_T("Error: Site redirects with no new location"));
          addMessage(message);
          pHttpFile->Close();
          delete pHttpFile;
          pHttpFile = 0;
          pHttpServer->Close();
          delete pHttpServer;
          pHttpServer = 0;
          return 2;
        }

        strNewLocation = strNewLocation.Mid(nPlace + 10);
        nPlace = strNewLocation.Find('\n');
        if(nPlace > 0)
        {
          strNewLocation = strNewLocation.Left(nPlace);
        }
        nPlace = strNewLocation.Find('\r');
        if(nPlace > 0)
        {
          strNewLocation = strNewLocation.Left(nPlace);
        }

        // close up the redirected site

        pHttpFile->Close();
        delete pHttpFile;
        pHttpFile = 0;
        pHttpServer->Close();
        delete pHttpServer;
        pHttpServer = 0;

        CString message;
        message.Format(_T("Redirected to %s"), (LPCTSTR)strNewLocation);
        addMessage(message);
        if(nWaitTime == 1)
        {
          message.Format(_T("Waiting %d %s"), nWaitTime, _T("second"));
        }
        else
        {
          message.Format(_T("Waiting %d %s"), nWaitTime, _T("seconds"));
        }
        addMessage(message);
        for(long n = 0; n < nWaitTime; n++)
        {
          if(bCancel)
          {
            addMessage(_T("Download canceled by user."));
			CgeoDataClientDialog* dialog = (CgeoDataClientDialog*)messagesListBox.GetParent();
			if(dialog)
			{
				dialog->proceedButton.EnableWindow();
				CWnd *pLeft = dialog->GetDlgItem(IDC_EDIT_LEFT);
				if(pLeft) pLeft->EnableWindow();
				CWnd *pTop = dialog->GetDlgItem(IDC_EDIT_TOP);
				if(pTop) pTop->EnableWindow();
			}
            return 0;
          }
          Sleep(1000);
        }

        strUrl = strNewLocation; // Start over.

        break;
      }
      case HTTP_STATUS_OK:
      {
        CString strNewLocation;
        int nWaitTime = 0;
        pHttpFile->QueryInfo(HTTP_QUERY_RAW_HEADERS_CRLF, strNewLocation);
        int nPlace = strNewLocation.Find(_T("Refresh: "));
        if(nPlace != -1)
        {
          strNewLocation = strNewLocation.Mid(nPlace + 9);
          nPlace = strNewLocation.Find('\n');
          if(nPlace > 0)
          {
            strNewLocation = strNewLocation.Left(nPlace);
          }
          nPlace = strNewLocation.Find('\r');
          if(nPlace > 0)
          {
            strNewLocation = strNewLocation.Left(nPlace);
            // Example: "10; URL=/diststatus/servlet/gov.usgs.edc.RequestStatus?zid=20070320.042438789.212041245002"
          }
        }
        else
        {
          strNewLocation.Empty();
        }
        nPlace = strNewLocation.Find(_T(";"));
        if(nPlace > 0)
        {
          CString strWaitTime = strNewLocation.Left(nPlace);
          nWaitTime = atoi(strWaitTime);
          strNewLocation = strNewLocation.Mid(nPlace + 1);
          // Example: " URL=/diststatus/servlet/gov.usgs.edc.RequestStatus?zid=20070320.042438789.212041245002"
          int httpFound = strNewLocation.Find(_T("http://"));
          if(httpFound > 0)
          {
            // The HTTP string is already present in the redirected URL (full path redirection).
            strNewLocation = strNewLocation.Mid(httpFound);
          }
          else
          {
            nPlace = strNewLocation.Find(_T("="));
            if(nPlace > 0)
            {
              strNewLocation = "http://" + strServerName + strNewLocation.Mid(nPlace + 1);
              // Example: "/diststatus/servlet/gov.usgs.edc.RequestStatus?zid=20070320.042438789.212041245002"
            }
          }
          // Download the workload (the HTML file) and
          // search for text worth displaying:
          // The progress information text.
          // Such an HTML file typically displays a
          // "Please wait for the download to be finished..." message.
          data.Empty();
          TCHAR sz[BLOCKSIZE];
          int nRead = pHttpFile->Read(sz, BLOCKSIZE - 1);
          while(nRead > 0)
          {
            sz[nRead] = '\0';
            data += sz;
            nRead = pHttpFile->Read(sz, BLOCKSIZE - 1);
          }
          POSITION pos = progressInformationText.GetHeadPosition();
          for(int i = 0; i < progressInformationText.GetCount(); i++)
          {
            CString string = progressInformationText.GetNext(pos);
            if(data.Find(string) != -1)
            {
              // Text found. Display.
              addHTMLMessage(data);
              break;
            }
          }
          if(i == progressInformationText.GetCount())
          {
            // No text found at all. Write to log file.
            CStdioFile logFile;
            if(logFile.Open("geoDataClient.log", CFile::modeCreate | CFile::modeNoTruncate | CFile::modeWrite))
            {
              CString text;
              CTime now = CTime::GetCurrentTime();
              text = now.FormatGmt("geoDataClientDialog log file created %A, %B %d, %Y %H:%M:%S UTC\r\n");
              logFile.WriteString(text);
              logFile.WriteString(_T("\r\n"));
              logFile.WriteString(_T("CHTTPRequest::execute():\r\n"));
              logFile.WriteString(_T("no progressInformationText found.\r\n"));
              logFile.WriteString(_T("Dumping whole HTTP response body:\r\n"));
              logFile.WriteString(_T("\r\n"));
              logFile.WriteString(_T("+++ BEGIN OF PAYLOAD +++\r\n"));
              logFile.WriteString(data);
              logFile.WriteString(_T("+++ END OF PAYLOAD +++\r\n"));
              logFile.WriteString(_T("\r\n"));
              logFile.Close();
            }
          }
        }

        // close up

        if(strNewLocation.IsEmpty())
        {
          if(bMemoryOnly)
          {
            // The status is OK. No output filename specified.
            // Download the result data to the CString variable "data".
            data.Empty();
            TCHAR sz[BLOCKSIZE];
            int nRead = pHttpFile->Read(sz, BLOCKSIZE - 1);
            while(nRead > 0)
            {
              sz[nRead] = '\0';
              data += sz;
              nRead = pHttpFile->Read(sz, BLOCKSIZE - 1);
            }
            message.Format(_T("Download of \"%s\" finished."), strOriginalUrl);
            addMessage(message);
          }
          else
          {
            if(suggestedOutputFilename.IsEmpty())
            {
              outputFilename = pHttpFile->GetFileName();
            }
            else
            {
              outputFilename = suggestedOutputFilename;
            }
            char fullPath[MAX_PATH];
            char *lpFilePart;
            DWORD dwResult =
              GetFullPathName(outputFilename,
                              MAX_PATH,
                              fullPath,
                              &lpFilePart);
            outputFilename = fullPath;

            CString strContentLength;
            pHttpFile->QueryInfo(HTTP_QUERY_CONTENT_LENGTH, strContentLength);
            DWORD contentLength = atol(strContentLength);
            DWORD contentLengthSize = 0;
            progressControl.SetRange(0, 100);

            // The status is OK. An output filename has been specified.
            // Download the result data to the output file.
            message.Format(_T("Writing file \"%s\"..."), outputFilename);
            addMessage(message);
            DWORD bytesRead = 0;
            FILE *outputFile;
            outputFile = fopen(outputFilename, "wb");
            if(outputFile)
            {
              TCHAR sz[BLOCKSIZE];
              UINT nRead = pHttpFile->Read(sz, BLOCKSIZE - 1);
              while(nRead > 0)
              {
                sz[nRead] = '\0';
                fwrite(sz, 1, nRead, outputFile);
                bytesRead += nRead;
                UINT percent = UINT(100.0 * double(bytesRead) / double(contentLength) + 0.5);
                progressControl.SetPos(percent);
                nRead = pHttpFile->Read(sz, BLOCKSIZE - 1);
                if(bCancel)
                {
                  fclose(outputFile);
                  DeleteFile(outputFilename);
                  addMessage(_T("Download canceled by user."));
                  pHttpFile->Close();
                  delete pHttpFile;
                  pHttpFile = 0;
                  pHttpServer->Close();
                  delete pHttpServer;
                  pHttpServer = 0;
					CgeoDataClientDialog* dialog = (CgeoDataClientDialog*)messagesListBox.GetParent();
					if(dialog)
					{
						dialog->proceedButton.EnableWindow();
						CWnd *pLeft = dialog->GetDlgItem(IDC_EDIT_LEFT);
						if(pLeft) pLeft->EnableWindow();
						CWnd *pTop = dialog->GetDlgItem(IDC_EDIT_TOP);
						if(pTop) pTop->EnableWindow();
					}
                  return 3;
                }
              }
              fclose(outputFile);
              progressControl.SetPos(0);
            }
            else
            {
              message.Format(_T("Error: Writing file \"%s\" failed."), outputFilename);
              addMessage(message);
              pHttpFile->Close();
              delete pHttpFile;
              pHttpFile = 0;
              pHttpServer->Close();
              delete pHttpServer;
              pHttpServer = 0;
              return 4;
            }
            message.Format(_T("Writing file \"%s\" finished."), outputFilename);
            addMessage(message);
          }
          pHttpFile->Close();
          delete pHttpFile;
          pHttpFile = 0;
          pHttpServer->Close();
          delete pHttpServer;
          pHttpServer = 0;
          return 0; // This is the only control path that returns OK.
        }
        else
        {
          // The status is OK. A new URL is present.
          // Repeat the HTTP request there.
          pHttpFile->Close();
          delete pHttpFile;
          pHttpFile = 0;
          pHttpServer->Close();
          delete pHttpServer;
          pHttpServer = 0;

          CString message;
          message.Format(_T("Redirected to %s"), (LPCTSTR)strNewLocation);
          addMessage(message);
          if(nWaitTime == 1)
          {
            message.Format(_T("Waiting %d %s"), nWaitTime, _T("second"));
          }
          else
          {
            message.Format(_T("Waiting %d %s"), nWaitTime, _T("seconds"));
          }
          addMessage(message);
          for(long n = 0; n < nWaitTime; n++)
          {
            if(bCancel)
            {
              addMessage(_T("Download canceled by user."));
				CgeoDataClientDialog* dialog = (CgeoDataClientDialog*)messagesListBox.GetParent();
				if(dialog)
				{
					dialog->proceedButton.EnableWindow();
					CWnd *pLeft = dialog->GetDlgItem(IDC_EDIT_LEFT);
					if(pLeft) pLeft->EnableWindow();
					CWnd *pTop = dialog->GetDlgItem(IDC_EDIT_TOP);
					if(pTop) pTop->EnableWindow();
				}
              return 3;
            }
            Sleep(1000);
          }

          strUrl = strNewLocation; // Start over.
        }
        break;
      }
      default:
      {
        message.Format(_T("Error: Status Code is %lu\n"), dwRet);
        addMessage(message);
        CString strResponseHeaders;
        pHttpFile->QueryInfo(HTTP_QUERY_RAW_HEADERS_CRLF, strResponseHeaders);
        message.Format(_T("Response Headers: \"%s\""), strResponseHeaders);
        addMessage(message);
        pHttpFile->Close();
        delete pHttpFile;
        pHttpFile = 0;
        pHttpServer->Close();
        delete pHttpServer;
        pHttpServer = 0;
        return 5;
      }
    }
  }
}

CString CHTTPRequest::getData()
{
  return data;
}

void CHTTPRequest::addMessage(const CString& text)
{
  CString resToken;
  int curPos = 0;
  resToken = text.Tokenize("\n\r", curPos);
  if(resToken.Find("Response Headers:") == 0)
  {
    addMessageOneLine("Response Headers:");
    resToken = resToken.Mid(19);
    while(resToken != "")
    {
      if((resToken != "") && (resToken != "\""))
      {
        resToken = "  " + resToken;
        addMessageOneLine(resToken);
      }
      resToken = text.Tokenize("\n\r",curPos);
    }
  }
  while(resToken != "")
  {
    if(resToken != "")
    {
      addMessageOneLine(resToken);
    }
    resToken = text.Tokenize("\n\r",curPos);
  }
}

void CHTTPRequest::addMessageOneLine(const CString& text)
{
  CDC* pDC = messagesListBox.GetParent()->GetWindowDC();
  HFONT hFont = (HFONT)messagesListBox.GetParent()->SendMessage(WM_GETFONT);
  CFont *pFont = CFont::FromHandle(hFont);
  ASSERT(pFont);
  CFont* pPrevFont = pDC->SelectObject(pFont);

  CSize newExtent = pDC->GetTextExtent(text, text.GetLength());
  if(newExtent.cx > maxListboxTextExtent)
  {
    maxListboxTextExtent = newExtent.cx;
  }

  messagesListBox.SetHorizontalExtent(maxListboxTextExtent + 6);
  pDC->SelectObject(pPrevFont);
  messagesListBox.GetParent()->ReleaseDC(pDC);

  messagesListBox.AddString(text);
  int position = messagesListBox.GetCount() - 1;
  messagesListBox.SetCurSel(position);
}

void CHTTPRequest::addHTMLMessage(const CString& text)
{
  int beginStrong = FindCaseInsensitive(text, "<STRONG>");
  if(beginStrong == -1)
  {
    int beginH2 = FindCaseInsensitive(text, "<H2>");
    if(beginH2 == -1)
    {
      addToList(text);
    }
    else
    {
      while(beginH2 != -1)
      {
        int endH2 = FindCaseInsensitive(text, "</H2>");
        if(endH2 != -1)
        {
          CString tokenH2 = text.Mid(beginH2 + 4, endH2 - beginH2 - 4);
          int posBR = FindCaseInsensitive(tokenH2, "<BR>");
          while(posBR != -1)
          {
            CString tokenBR;
            if(posBR != -1)
            {
              tokenBR = tokenH2.Left(posBR);
              tokenH2 = tokenH2.Mid(posBR + 4);
            }
            else
            {
              tokenBR = tokenH2;
              tokenH2.Empty();
            }
            addToList(tokenBR);
            posBR = FindCaseInsensitive(tokenH2, "<BR>");
          }
          if(tokenH2.GetLength() != 0)
          {
            addToList(tokenH2);
          }
        }
        beginH2 = FindCaseInsensitive(text, "<H2>", beginH2 + 1);
      }
    }
  }
  else
  {
    while(beginStrong != -1)
    {
      int endStrong = FindCaseInsensitive(text, "</STRONG>");
      if(endStrong != -1)
      {
        CString tokenSTRONG = text.Mid(beginStrong + 8, endStrong - beginStrong - 8);
        int posBR = FindCaseInsensitive(tokenSTRONG, "<BR>");
        while(posBR != -1)
        {
          CString tokenBR;
          if(posBR != -1)
          {
            tokenBR = tokenSTRONG.Left(posBR);
            tokenSTRONG = tokenSTRONG.Mid(posBR + 4);
          }
          else
          {
            tokenBR = tokenSTRONG;
            tokenSTRONG.Empty();
          }
          addToList(tokenBR);
          posBR = FindCaseInsensitive(tokenSTRONG, "<BR>");
        }
        if(tokenSTRONG.GetLength() != 0)
        {
          addToList(tokenSTRONG);
        }
      }
      beginStrong = FindCaseInsensitive(text, "<STRONG>", beginStrong + 1);
    }
  }
}

void CHTTPRequest::addToList(CString string)
{
  int whitespace = string.FindOneOf("\t\r\n");
  while(whitespace != -1)
  {
    if(whitespace == 0)
    {
      string = string.Mid(1);
    }
    else
    {
      string = string.Left(whitespace - 1) + string.Mid(whitespace + 1);
    }
    whitespace = string.FindOneOf("\t\r\n");
  }
  if(string.GetLength() == 0)
  {
    return;
  }
  CDC* pDC = messagesListBox.GetParent()->GetWindowDC();
  HFONT hFont = (HFONT)messagesListBox.GetParent()->SendMessage(WM_GETFONT);
  CFont *pFont = CFont::FromHandle(hFont);
  ASSERT(pFont);
  CFont* pPrevFont = pDC->SelectObject(pFont);

  CSize newExtent = pDC->GetTextExtent(string, string.GetLength());
  if(newExtent.cx > maxListboxTextExtent)
  {
    maxListboxTextExtent = newExtent.cx;
  }

  messagesListBox.SetHorizontalExtent(maxListboxTextExtent + 6);
  pDC->SelectObject(pPrevFont);
  messagesListBox.GetParent()->ReleaseDC(pDC);

  messagesListBox.AddString(string);
  int position = messagesListBox.GetCount() - 1;
  messagesListBox.SetCurSel(position);
}

// Returns the zero based index to the first occurrence of strSearch in string,
// or -1 if strSearch does not appear in string.
// If strSearch points to a string of zero length, the function returns 0.
int CHTTPRequest::FindCaseInsensitive(const CString& string, const CString& stringSearch, int pos)
{
  CString string1 = string;
  string1.MakeLower();
  CString stringSearch1 = stringSearch;
  stringSearch1.MakeLower();
  return string1.Find(stringSearch1, pos);
}

CString CHTTPRequest::getOutputFilename()
{
  return outputFilename;
}
